import json
import logging
import subprocess
import shlex
import re

logging.basicConfig(level=logging.DEBUG)


class Tower:
    """
    A Python class that serves as a wrapper for 'tw' CLI commands.

    The class enables the execution of 'tw' commands in an object-oriented manner.
    You can call any subcommand of 'tw' as a method on instances of this class.
    The arguments of the subcommand can be passed as arguments to the method.

    Each command is run in a subprocess, with the output being captured and returned.
    """

    class TwCommand:
        def __init__(self, tw_instance, cmd):
            self.tw_instance = tw_instance
            self.cmd = cmd

        def __call__(self, *args, **kwargs):
            command = self.cmd.split()
            command.extend(args)
            return self.tw_instance._tw_run(command, **kwargs)

    # Constructs a new Tower instance with a specified workspace
    def __init__(self):
        pass

    # Executes a 'tw' command in a subprocess and returns the output.
    def _tw_run(self, cmd, *args, **kwargs):
        """
        Run a tw command with supplied commands
        """
        command = ["tw"]
        if kwargs.get("to_json"):
            to_json = True
            command.extend(["-o", "json"])
        else:
            to_json = False
        command.extend(cmd)
        command.extend(args)

        if kwargs.get("config") is not None:
            config_path = kwargs["config"]
            command.append(f"--config={config_path}")

        if "params_file" in kwargs:
            params_path = kwargs["params_file"]
            command.append(f"--params-file={params_path}")

        full_cmd = " ".join(
            arg if arg.startswith("$") else shlex.quote(arg) for arg in command
        )
        logging.debug(f" Running command: {full_cmd}\n")

        # Run the command and return the stdout
        process = subprocess.Popen(
            full_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True
        )
        stdout, _ = process.communicate()
        stdout = stdout.decode("utf-8").strip()

        # Error handling for stdout
        if stdout:
            if re.search(r"ERROR: .* already exists", stdout):
                raise ResourceExistsError(
                    " Resource already exists and will not be created."
                    " Please set 'overwrite: true'\n"
                )
            elif re.search(r"ERROR: .*", stdout):
                raise ResourceCreationError(
                    f" Resource creation failed with the following error: '{stdout}'.\n"
                    "Please check your config file and try again.\n"
                )
            elif to_json is True:
                return json.loads(stdout)
            else:
                return stdout

    # Allow any 'tw' subcommand to be called as a method.
    def __getattr__(self, cmd):
        """
        Magic method to allow any 'tw' subcommand to be called as a method.
        Returns a TwCommand object that can be called with arguments.
        """
        cmd = cmd.replace("_", "-")  # replace underscores with hyphens
        return self.TwCommand(self, cmd)


class ResourceExistsError(Exception):
    pass


class ResourceCreationError(Exception):
    pass



import logging

# Import the tw-pywrap package
from tw_pywrap import tower

logging.basicConfig(level=logging.DEBUG)

# Construct a new tw-pywrap Tower instance
tw = tower.Tower()

# Customise the entries below as required
workspace = "<YOUR_WORKSPACE>"  # Name of your Workspace
compute_env = "<YOUR_COMPUTE_ENVIRONMENT>"  # Name of your Compute Environment

# Specify a human-readable run name
run_name = "hello-world-tw-pywrap"

# Launch the 'hello-world' pipeline using the 'launch' method
pipeline_run = tw.launch(
    "--workspace",
    workspace,
    "--compute-env",
    compute_env,
    "--name",
    run_name,
    "--revision",
    "master",
    "--wait",
    "SUBMITTED",
    "https://github.com/nextflow-io/hello",
    to_json=True,
)



"""
Subclass of Tower class for overriding pipelines subcommand methods.
"""
from pathlib import Path
from tw_py.tower import Tower


class Pipelines(Tower):
    """
    Python wrapper for tw pipelines export command.
    """

    def export_pipeline(self, name, *args, **kwargs):
        """
        Export a pipeline
        """
        # create a Path object for the workspace directory
        workspace_dir = Path(self.workspace)

        # create the directory if it doesn't exist
        workspace_dir.mkdir(parents=True, exist_ok=True)

        # define the output file path
        outfile = str(workspace_dir / f"{name}.json")

        # Build the command
        command = [
            "pipelines",
            "export",
            "--workspace",
            self.workspace,
            "--name",
            name,
            outfile,
        ]

        # Pass the built command to the base class method in Tower
        return self._tw_run(command, *args, **kwargs)



"""
Subclass of Tower class for overriding compute environments subcommand methods.
"""
from pathlib import Path
from .tower import Tower


class ComputeEnvs(Tower):
    """
    Python wrapper for tw compute-envs export command.
    """

    def export_ce(self, name, *args, **kwargs):
        """
        Export a compute environment
        """
        # create a Path object for the workspace directory
        workspace_dir = Path(self.workspace)

        # create the directory if it doesn't exist
        workspace_dir.mkdir(parents=True, exist_ok=True)

        # define the output file path
        outfile = workspace_dir / f"{name}.json"

        # Build the command
        command = [
            "compute-envs",
            "export",
            "--workspace",
            self.workspace,
            "--name",
            name,
            str(outfile),
        ]

        # Pass the built command to the base class method in Tower
        return self._tw_run(command, *args, **kwargs, to_json=True)



"""
This script is used to build a Tower instance from a YAML configuration file.
Requires a YAML file that defines the resources to be created in Tower and
the required options for each resource based on the Tower CLI.
"""
import argparse
import logging
import time
import yaml

from pathlib import Path
from tw_pywrap import tower, helper, overwrite
from tw_pywrap.tower import ResourceCreationError, ResourceExistsError


logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--log_level",
        default="INFO",
        choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
        help="The desired log level (default: INFO).",
        type=str.upper,
    )
    parser.add_argument(
        "yaml", type=Path, help="Config file with Tower resources to create"
    )
    return parser.parse_args()


class BlockParser:
    """
    Manages blocks of commands defined in a configuration file and calls appropriate
    functions for each block for custom handling of command-line arguments to _tw_run().
    """

    def __init__(self, tw, list_for_add_method):
        """
        Initializes a BlockParser instance.

        Args:
        tw: A Tower class instance.
        list_for_add_method: A list of blocks that need to be
        handled by the 'add' method.
        """
        self.tw = tw
        self.list_for_add_method = list_for_add_method
        # Create an instance of Overwrite class
        self.overwrite_method = overwrite.Overwrite(self.tw)

    def handle_block(self, block, args):
        # Handles a block of commands by calling the appropriate function.
        block_handler_map = {
            "teams": (helper.handle_teams),
            "participants": (helper.handle_participants),
            "compute-envs": lambda tw, args: helper.handle_generic_block(
                tw, "compute_envs", args, method_name="import"
            ),
            "pipelines": (helper.handle_pipelines),
            "launch": lambda tw, args: helper.handle_generic_block(
                tw, "launch", args, method_name=None
            ),
        }

        # Check if overwrite is set to True, and call overwrite handler
        overwrite_option = args.get("overwrite", False)
        if overwrite_option:
            logging.debug(f" Overwrite is set to 'True' for {block}\n")
            self.overwrite_method.handle_overwrite(
                block, args["cmd_args"], overwrite_option
            )
        else:
            self.overwrite_method.handle_overwrite(block, args["cmd_args"])

        if block in self.list_for_add_method:
            helper.handle_generic_block(self.tw, block, args["cmd_args"])
        elif block in block_handler_map:
            block_handler_map[block](self.tw, args["cmd_args"])
        else:
            logger.error(f"Unrecognized resource block in YAML: {block}")


def main():
    options = parse_args()
    logging.basicConfig(level=options.log_level)

    tw = tower.Tower()
    block_manager = BlockParser(
        tw,
        [
            "organizations",  # all use method.add
            "workspaces",
            "credentials",
            "secrets",
            "actions",
            "datasets",
        ],
    )
    try:
        with open(options.yaml, "r") as f:
            data = yaml.safe_load(f)

        # Returns a dict that maps block names to lists of command line arguments.
        cmd_args_dict = helper.parse_all_yaml(options.yaml, list(data.keys()))

        for block, args_list in cmd_args_dict.items():
            for args in args_list:
                try:
                    # Run the 'tw' methods for each block
                    block_manager.handle_block(block, args)
                    time.sleep(3)
                except ResourceExistsError as e:
                    logging.error(e)
                    continue
    except ResourceCreationError as e:
        logging.error(e)


if __name__ == "__main__":
    main()



import json
import tempfile
import yaml
from urllib.parse import urlparse


def find_key_value_in_dict(data, target_key, target_value, return_key):
    """
    Generic method to find a key-value pair in a nested dictionary and within
    lists of dictionaries.
    Can use the input of json.loads() converting JSON string to dict.
    #TODO: huge candidate for refactoring
    """
    if isinstance(data, dict):
        for key, value in data.items():
            if key == target_key and value == target_value:
                # If we find our target key-value pair,
                # return the value of the specified return key if it is not None
                if return_key is not None:
                    return data.get(return_key)
                else:
                    # or any other value to indicate the existence of the key-value pair
                    return True
            if isinstance(value, dict):
                result = find_key_value_in_dict(
                    value, target_key, target_value, return_key
                )
                if result:
                    return result
            if isinstance(value, list):
                for item in value:
                    result = find_key_value_in_dict(
                        item, target_key, target_value, return_key
                    )
                    if result:
                        return result
    elif isinstance(data, list):
        for item in data:
            result = find_key_value_in_dict(item, target_key, target_value, return_key)
            if result:
                return result
    return None


def check_if_exists(json_data, namekey, namevalue):
    """
    Wrapper around find_key_value_in_dict() to validate that a resource was
    created successfully in Tower by looking for the name and value.
    """
    data = json.loads(json_data)
    if find_key_value_in_dict(data, namekey, namevalue, return_key=None):
        return True


def is_valid_yaml(file_path):
    """
    Check if a file is valid YAML
    """
    try:
        with open(file_path, "r") as file:
            yaml.safe_load(file)
        return True
    except yaml.YAMLError:
        return False


def get_pipeline_repo(pipeline_repo):
    """
    Provide full nf-core github URL if
    only the repo name is provided
    """
    if "nf-core" in pipeline_repo.lower():
        return "https://github.com/" + pipeline_repo
    else:
        return pipeline_repo


def is_url(s):
    """
    Check if a string is a valid URL
    """
    try:
        result = urlparse(s)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def create_temp_yaml(params_dict):
    """
    Create a generic temporary yaml file given a dictionary
    """

    class quoted_str(str):
        pass

    def quoted_str_representer(dumper, data):
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="'")

    yaml.add_representer(quoted_str, quoted_str_representer)
    for k, v in params_dict.items():
        if k == "outdir" and isinstance(v, str):
            params_dict[k] = quoted_str(v)

    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".yaml"
    ) as temp_file:
        yaml.dump(params_dict, temp_file)
        return temp_file.name



"""
This file contains helper functions for the library.
Including handling methods for each block in the YAML file, and parsing
methods for each block in the YAML file.
"""
import yaml
from tw_pywrap import utils


def parse_yaml_block(file_path, block_name):
    # Load the YAML file.
    with open(file_path, "r") as f:
        data = yaml.safe_load(f)

    # Get the specified block.
    block = data.get(block_name)

    # Initialize an empty list to hold the lists of command line arguments.
    cmd_args_list = []

    # Initialize a set to track the --name values within the block.
    name_values = set()

    # Iterate over each item in the block.
    for item in block:
        cmd_args = parse_block(block_name, item)
        name = find_name(cmd_args)
        if name in name_values:
            raise ValueError(
                f" Duplicate 'name' specified in config file"
                f" for {block_name}: {name}. Please specify a unique name."
            )
        name_values.add(name)

        cmd_args_list.append(cmd_args)

    # Return the block name and list of command line argument lists.
    return block_name, cmd_args_list


def parse_all_yaml(file_path, block_names):
    resource_order = [
        "organizations",
        "teams",
        "workspaces",
        "participants",
        "credentials",
        "compute-envs",
        "secrets",
        "actions",
        "datasets",
        "pipelines",
        "launch",
    ]
    # Initialize an empty dictionary to hold all the command arguments.
    cmd_args_dict = {}

    # Iterate over each block name in the desired order.
    for block_name in resource_order:
        # Check if the block name is present in the provided block_names list
        if block_name in block_names:
            # Parse the block and add its command arguments to the dictionary.
            block_name, cmd_args_list = parse_yaml_block(file_path, block_name)
            cmd_args_dict[block_name] = cmd_args_list

    # Return the dictionary of command arguments.
    return cmd_args_dict


def parse_block(block_name, item):
    # Define the mapping from block names to functions.
    block_to_function = {
        "credentials": parse_credentials_block,
        "compute-envs": parse_compute_envs_block,
        "teams": parse_teams_block,
        "actions": parse_actions_block,
        "datasets": parse_datasets_block,
        "pipelines": parse_pipelines_block,
        "launch": parse_launch_block,
    }
    # Use the generic block function as a default.
    parse_fn = block_to_function.get(block_name, parse_generic_block)
    overwrite = item.pop("overwrite", False)

    # Call the appropriate function and return its result along with overwrite value.
    cmd_args = parse_fn(item)
    return {"cmd_args": cmd_args, "overwrite": overwrite}


# Parsers for certain blocks of yaml that require handling
# for structuring command line arguments in a certain way


def parse_generic_block(item):
    cmd_args = []
    for key, value in item.items():
        cmd_args.extend([f"--{key}", str(value)])
    return cmd_args


def parse_credentials_block(item):
    cmd_args = []
    for key, value in item.items():
        if key == "type":
            cmd_args.append(str(value))
        else:
            cmd_args.extend([f"--{key}", str(value)])
    return cmd_args


def parse_compute_envs_block(item):
    cmd_args = []
    for key, value in item.items():
        if key == "file-path":
            cmd_args.append(str(value))
        else:
            cmd_args.extend([f"--{key}", str(value)])
    return cmd_args


def parse_teams_block(item):
    # Keys for each list
    cmd_keys = ["name", "organization", "description"]
    members_keys = ["name", "organization", "members"]

    cmd_args = []
    members_cmd_args = []

    for key, value in item.items():
        if key in cmd_keys:
            cmd_args.extend([f"--{key}", str(value)])
        elif key in members_keys and key == "members":
            for member in value:
                members_cmd_args.append(
                    [
                        "--team",
                        str(item["name"]),
                        "--organization",
                        str(item["organization"]),
                        "add",
                        "--member",
                        str(member),
                    ]
                )
    return (cmd_args, members_cmd_args)


def parse_actions_block(item):
    cmd_args = []
    temp_file_name = None
    for key, value in item.items():
        if key == "type":
            cmd_args.append(str(value))
        elif key == "params":
            temp_file_name = utils.create_temp_yaml(value)
            cmd_args.extend(["--params-file", temp_file_name])
        else:
            cmd_args.extend([f"--{key}", str(value)])
    return cmd_args


def parse_datasets_block(item):
    cmd_args = []
    for key, value in item.items():
        if key == "file-path":
            cmd_args.extend(
                [
                    str(item["file-path"]),
                    "--name",
                    str(item["name"]),
                    "--workspace",
                    str(item["workspace"]),
                    "--description",
                    str(item["description"]),
                ]
            )
        if key == "header" and value is True:
            cmd_args.append("--header")
    return cmd_args


def parse_pipelines_block(item):
    cmd_args = []
    repo_args = []
    params_args = []

    for key, value in item.items():
        if key == "url":
            repo_args.extend([str(value)])
        elif key == "params":
            temp_file_name = utils.create_temp_yaml(value)
            params_args.extend(["--params-file", temp_file_name])
        elif key == "file-path":
            repo_args.extend([str(value)])
        else:
            cmd_args.extend([f"--{key}", str(value)])

    # First append the url related arguments then append the remaining ones
    cmd_args = repo_args + params_args + cmd_args
    return cmd_args


def parse_launch_block(item):
    repo_args = []
    cmd_args = []
    for key, value in item.items():
        if key == "pipeline" or key == "url":
            repo_args.extend([str(value)])
        elif key == "params":
            temp_file_name = utils.create_temp_yaml(value)
            cmd_args.extend(["--params-file", temp_file_name])
        else:
            cmd_args.extend([f"--{key}", str(value)])
    cmd_args = repo_args + cmd_args

    return cmd_args


# Handlers to call the actual tower method,based on the block name.
# Certain blocks required special handling and combination of methods.


def handle_generic_block(tw, block, args, method_name="add"):
    # Generic handler for most blocks, with optional method name
    method = getattr(tw, block)
    if method_name is None:
        method(*args)
    else:
        method(method_name, *args)


def handle_teams(tw, args):
    cmd_args, members_cmd_args = args
    tw.teams("add", *cmd_args)
    for sublist in members_cmd_args:
        tw.teams("members", *sublist)


def handle_participants(tw, args):
    # Generic handler for blocks with a key to skip
    method = getattr(tw, "participants")
    skip_key = "--role"
    new_args = [
        arg
        for i, arg in enumerate(args)
        if not (args[i - 1] == skip_key or arg == skip_key)
    ]
    method("add", *new_args)
    method("update", *args)


def handle_pipelines(tw, args):
    method = getattr(tw, "pipelines")
    for arg in args:
        # Check if arg is a url or a json file.
        # If it is, use the appropriate method and break.
        if utils.is_url(arg):
            method("add", *args)
            break
        elif ".json" in arg:
            method("import", *args)


def find_name(cmd_args):
    """
    Find and return the value associated with --name in the cmd_args list.
    """
    args_list = cmd_args.get("cmd_args", [])
    for i in range(len(args_list) - 1):
        if args_list[i] == "--name":
            return args_list[i + 1]
    return None



import unittest
from unittest.mock import patch
from tw_pywrap import tower
import json
import subprocess


class TestTower(unittest.TestCase):
    def setUp(self):
        self.tw = tower.Tower()

    @patch("subprocess.Popen")
    def test_run_with_jsonout_command(self, mock_subprocess):
        mock_pipelines_json = {
            "id": "5lWcpupLHnHkq9fM5JYaOn",
            "computeEnvId": "403VpC7AetAmj42MnMOAwJ",
            "pipeline": "https://github.com/nextflow-io/hello",
            "workDir": "s3://myworkdir/test",
            "revision": "",
            "configText": "",
            "paramsText": "",
            "resume": "false",
            "pullLatest": "false",
            "stubRun": "false",
            "dateCreated": "2023-02-15T13:14:30Z",
        }
        # Mock the stdout of the Popen process
        mock_subprocess.return_value.communicate.return_value = (
            json.dumps(mock_pipelines_json).encode(),
            b"",
        )

        # Dynamically get the pipelines command
        command = getattr(self.tw, "pipelines")

        # Run the command with arguments
        result = command("view", "--name", "pipeline_name", to_json=True)

        # Check that Popen was called with the right arguments
        mock_subprocess.assert_called_once_with(
            "tw -o json pipelines view --name pipeline_name",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
        )

        # Check that the output was decoded correctly
        self.assertEqual(result, mock_pipelines_json)

    def test_resource_exists_error(self):
        with patch("subprocess.Popen") as mock_subprocess:
            # Simulate a 'resource already exists' error
            mock_subprocess.return_value.communicate.return_value = (
                b"ERROR: Resource already exists",
                b"",
            )

            command = getattr(self.tw, "pipelines")

            # Check that the error is raised
            with self.assertRaises(tower.ResourceExistsError):
                command("arg1", "arg2")

    def test_resource_creation_error(self):
        with patch("subprocess.Popen") as mock_subprocess:
            # Simulate a 'resource creation failed' error
            mock_subprocess.return_value.communicate.return_value = (
                b"ERROR: Resource creation failed",
                b"",
            )

            command = getattr(self.tw, "pipelines")

            # Check that the error is raised
            with self.assertRaises(tower.ResourceCreationError):
                command("import", "my_pipeline.json", "--name", "pipeline_name")

    def test_json_parsing(self):
        with patch("subprocess.Popen") as mock_subprocess:
            # Mock the stdout of the Popen process to return JSON
            mock_subprocess.return_value.communicate.return_value = (
                b'{"key": "value"}',
                b"",
            )

            command = getattr(self.tw, "pipelines")

            # Check that the JSON is parsed correctly
            self.assertEqual(command("arg1", "arg2", to_json=True), {"key": "value"})


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from setuptools import find_packages, setup

version = "0.1.0"

with open("README.md") as f:
    readme = f.read()

setup(
    name="tw-pywrap",
    version=version,
    description="""Automation of resource creation to be able 
    to setup and run pipelines on Nextflow Tower.""",
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords=["nextflow", "bioinformatics", "workflow", "pipeline", "nextflow-tower"],
    author="Esha Joshi, Adam Talbot, Harshil Patel",
    author_email="esha.joshi@seqera.io, adam.talbot@seqera.io, harshil.patel@seqera.io",
    url="https://github.com/seqeralabs/tw-pywrap",
    license="MIT",
    entry_points={"console_scripts": ["tw-pywrap=tw_pywrap.cli:main"]},
    python_requires=">=3.8, <4",  # untested
    install_requires=["pyyaml>=6.0.0"],
    packages=find_packages(exclude=("docs")),
    include_package_data=True,
    zip_safe=False,
)



import unittest
from unittest.mock import patch
from tw_pywrap import tower


class TestTowerSubcommands(unittest.TestCase):
    """
    Unit tests for the subcommands of the Tower base class.
    These tests check that the subcommands of the Tower base class
    construct the proper commands to be passed to the _tw_run method of the base class
    when dynamically called as methods of the Tower class.
    """

    def setUp(self):
        self.tw = tower.Tower()
        self.patcher = patch.object(self.tw, "_tw_run")
        self.mock_tw_run = self.patcher.start()
        # Make the mock behave like a function
        self.mock_tw_run.side_effect = lambda *args, **kwargs: "mocked result"

    def subcommand_test(self, subcommand, args, expected_command, **kwargs):
        with self.subTest(subcommand=subcommand):
            command = getattr(self.tw, subcommand)
            result = command(*args, **kwargs)
            self.mock_tw_run.assert_called_once_with(expected_command, **kwargs)
            self.assertEqual(
                result, "mocked result"
            )  # ensure the _tw_run mock returned the correct value
            self.mock_tw_run.reset_mock()  # Reset the mock for the next subcommand

    def test_list_command(self):
        """
        Unit test for the list method of Tower base class.
        This test checks that the list method constructs the
        proper commands to be passed to the _tw_run method of the base class.
        """

        resource_subcommands = [
            "pipelines",
            "compute_envs",
            "actions",
            "datasets",
            "credentials",
            "participants",
            "secrets",
            "runs",
        ]
        org_subcommands = ["teams", "members", "collaborators"]
        plain_subcommands = ["organizations", "workspaces"]

        list_workspace = ["list", "--workspace", "workspace_name"]
        list_organization = ["list", "--organization", "organization_name"]
        list_plain = ["list"]

        for subcommand in resource_subcommands:
            expected_command = (
                ["compute-envs", *list_workspace]
                if subcommand == "compute_envs"
                else [subcommand, *list_workspace]
            )
            self.subcommand_test(subcommand, list_workspace, expected_command)

        for subcommand in org_subcommands:
            self.subcommand_test(
                subcommand, list_organization, [subcommand, *list_organization]
            )

        for subcommand in plain_subcommands:
            self.subcommand_test(subcommand, list_plain, [subcommand, *list_plain])

    def test_add_pipelines_command(self):
        """
        Unit test for the add method of Pipelines class, overriding the base class.
        """
        add_command_with_config = [
            "pipelines",
            "add",
            "my_git_repo",
            "--name",
            "pipeline_name",
            "--workspace",
            "workspace_name",
        ]

        self.subcommand_test(
            "pipelines",
            [
                "add",
                "my_git_repo",
                "--name",
                "pipeline_name",
                "--workspace",
                "workspace_name",
            ],
            add_command_with_config,
            params_file="config_path",
        )

    def test_add_compute_envs_command(self):
        """
        Unit test for the add method of Compute Envs class, overriding the base class.
        """
        add_command = [
            "compute-envs",
            "add",
            "executor_name",
            "--name",
            "my_compute_env",
            "--workspace",
            "workspace_name",
            "--workdir",
            "s3://myworkdir",
        ]
        self.subcommand_test(
            "compute_envs",
            [
                "add",
                "executor_name",
                "--name",
                "my_compute_env",
                "--workspace",
                "workspace_name",
                "--workdir",
                "s3://myworkdir",
            ],
            add_command,
        )

    def test_delete_command(self):
        """
        Unit test for the delete method of Tower base class.
        """
        delete_command = [
            "pipelines",
            "delete",
            "pipeline_name",
            "--workspace",
            "workspace_name",
        ]
        self.subcommand_test(
            "pipelines",
            ["delete", "pipeline_name", "--workspace", "workspace_name"],
            delete_command,
        )

    def test_view_command(self):
        """
        Unit test for the view method of Tower base class.
        """
        view_command = [
            "pipelines",
            "view",
            "pipeline_name",
            "--workspace",
            "workspace_name",
        ]
        self.subcommand_test(
            "pipelines",
            ["view", "pipeline_name", "--workspace", "workspace_name"],
            view_command,
        )

    def test_import_command(self):
        """
        Unit test for the import method of Tower base class.
        """
        import_command = [
            "pipelines",
            "import",
            "pipeline_file.json",
            "--name",
            "my_pipeline_name",
            "--workspace",
            "workspace_name",
        ]
        self.subcommand_test(
            "pipelines",
            [
                "import",
                "pipeline_file.json",
                "--name",
                "my_pipeline_name",
                "--workspace",
                "workspace_name",
            ],
            import_command,
        )



import json
from tw_pywrap import utils
from tw_pywrap.tower import ResourceExistsError
import logging


class Overwrite:
    """
    Manages overwrite functionality for Tower resources.

    # TODO: When overwrite functionality becomes native to CLI,
    we can remove this class.
    """

    # Define blocks for simple overwrite with --name and --workspace
    generic_deletion = [
        "credentials",
        "secrets",
        "compute-envs",
        "datasets",
        "actions",
        "pipelines",
    ]

    def __init__(self, tw):
        """
        Initializes an Overwrite instance.

        Args:
        tw: A Tower class instance.

        Attributes:
        tw: A Tower class instance used to execute Tower CLI commands.
        cached_jsondata: A cached placeholder for JSON data. Default value is None.
        block_jsondata: A dictionary to store JSON data for each block.
        Key is the block name, and value is the corresponding JSON data.
        """
        self.tw = tw
        self.cached_jsondata = None
        self.block_jsondata = {}  # New dict to hold JSON data per block

        # Define special handlers for resources deleted with specific args
        self.block_operations = {
            "organizations": {
                "keys": ["name"],
                "method_args": self._get_organization_args,
                "name_key": "orgName",
            },
            "teams": {
                "keys": ["name", "organization"],
                "method_args": self._get_team_args,
                "name_key": "name",
            },
            "participants": {
                "keys": ["name", "type", "workspace"],
                "method_args": self._get_participant_args,
                "name_key": "email",
            },
            "workspaces": {
                "keys": ["name", "organization"],
                "method_args": self._get_workspace_args,
                "name_key": "workspaceId",
            },
        }

    def handle_overwrite(self, block, args, overwrite=False):
        """
        Handles overwrite functionality for Tower resources and
        calling the 'tw delete' method with the correct args.
        """
        if block in Overwrite.generic_deletion:
            self.block_operations[block] = {
                "keys": ["name", "workspace"],
                "method_args": self._get_generic_deletion_args,
                "name_key": "name",
            }

        if block in self.block_operations:
            operation = self.block_operations[block]
            keys_to_get = operation["keys"]
            self.cached_jsondata, tw_args = self._get_json_data(
                block, args, keys_to_get
            )

            if block == "participants":
                if tw_args.get("type") == "TEAM":
                    self.block_operations["participants"]["name_key"] = "teamName"
                else:
                    self.block_operations["participants"]["name_key"] = "email"

            if self.check_resource_exists(operation["name_key"], tw_args):
                # if resource exists, delete
                if overwrite:
                    logging.debug(
                        f" The attempted {block} resource already exists."
                        " Overwriting.\n"
                    )
                    self._delete_resource(block, operation, tw_args)
                else:  # return an error if resource exists, overwrite=False
                    raise ResourceExistsError(
                        f" The {block} resource already exists and"
                        " will not be created. Please set 'overwrite: True'"
                        " in your config file.\n"
                    )

    def _get_organization_args(self, args):
        """
        Returns a list of arguments for the delete() method for organizations.
        """
        return ("delete", "--name", args["name"])

    def _get_team_args(self, args):
        """
        Returns a list of arguments for the delete() method for teams. The teamId
        used to delete will be retrieved using the find_key_value_in_dict() method.
        """
        jsondata = self.block_jsondata.get("teams", None)

        if not jsondata:
            json_method = getattr(self.tw, "-o json")
            json_out = json_method("teams", "list", "-o", args["organization"])
            self.block_jsondata["teams"] = json_out
        else:
            json_out = jsondata

        # Get the teamId from the json data
        team_id = utils.find_key_value_in_dict(
            json.loads(json_out), "name", args["name"], "teamId"
        )
        return ("delete", "--id", str(team_id), "--organization", args["organization"])

    def _get_participant_args(self, args):
        """
        Returns a list of arguments for the delete() method for participants.
        Works for both teams and members where if the type is TEAM, the teamName
        is used to delete, and if the type is MEMBER, the email is used to delete.
        """
        method_args = (
            "delete",
            "--name",
            args["name"],
            "--type",
            args["type"],
            "--workspace",
            args["workspace"],
        )
        return method_args

    def _get_workspace_args(self, args):
        """
        Returns a list of arguments for the delete() method for workspaces. The
        workspaceId used to delete will be retrieved using the _find_workspace_id()
        method.
        """
        workspace_id = self._find_workspace_id(
            json.loads(self.tw.workspaces.list("-o", "json")),
            args["organization"],
            args["name"],
        )
        return ("delete", "--id", str(workspace_id))

    def _get_generic_deletion_args(self, args):
        """
        Returns a list of arguments for the delete() method for resources that
        are deleted with --name and --workspace.
        """
        return (
            "delete",
            "--name",
            args["name"],
            "--workspace",
            args["workspace"],
        )

    def _get_json_data(self, block, args, keys_to_get):
        """
        For the specified keys in the operations dictionary, get the values from
        the command line arguments and return a dictionary of values.

        Also, gets json data from Tower by calling the list() method once and caches
        the json data for that block in self.block_jsondata. If the block data already
        exists, it will be retrieved from the dictionary instead of calling the list().

        Returns a tuple of json data and a dictionary of values to run delete() on.
        """
        json_method = getattr(self.tw, "-o json")
        tw_args = []  # Default value for tw_args

        # Check if block data already exists
        if block in self.block_jsondata:
            self.cached_jsondata = self.block_jsondata[block]
            tw_args = self._get_values_from_cmd_args(args, keys_to_get)
        else:
            # Fetch the data if it does not exist
            if block == "teams":
                tw_args = self._get_values_from_cmd_args(args[0], keys_to_get)
                self.cached_jsondata = json_method(
                    block, "list", "-o", tw_args["organization"]
                )
            elif block in Overwrite.generic_deletion or block == "participants":
                tw_args = self._get_values_from_cmd_args(args, keys_to_get)
                self.cached_jsondata = json_method(
                    block, "list", "-w", tw_args["workspace"]
                )
            else:
                tw_args = self._get_values_from_cmd_args(args, keys_to_get)
                self.cached_jsondata = json_method(block, "list")

        # Store this data in the block_jsondata dict for later use
        self.block_jsondata[block] = self.cached_jsondata
        return self.cached_jsondata, tw_args

    def check_resource_exists(self, name_key, tw_args):
        """
        Check if a resource exists in Tower by looking for the name and value
        in the json data generated from the list() method.
        """
        return utils.check_if_exists(self.cached_jsondata, name_key, tw_args["name"])

    def _delete_resource(self, block, operation, tw_args):
        """
        Delete a resource in Tower by calling the delete() method and
        arguments defined in the operation dictionary.
        """
        method_args = operation["method_args"](tw_args)
        method = getattr(self.tw, block)
        method(*method_args)

    def _get_values_from_cmd_args(self, cmd_args, keys):
        """
        Return a dictionary of values from a list of command line arguments based
        on a input list of keys.
        """
        values = {key: None for key in keys}
        key = None

        for arg in cmd_args:
            if arg.startswith("--"):
                key = arg[2:]
            else:
                if key and key in keys:
                    values[key] = arg
                key = None
        return values

    def _find_workspace_id(self, organization, workspace_name):
        """
        Custom method to find a workspace ID in a nested dictionary with a given
        organization name and workspace name. This ID will be used to delete the
        workspace.
        """
        if "workspaces" in self.cached_jsondata:
            workspaces = self.cached_jsondata["workspaces"]
            for workspace in workspaces:
                if (
                    workspace.get("orgName") == organization
                    and workspace.get("workspaceName") == workspace_name
                ):
                    return workspace.get("workspaceId")
        return None
